"""BiTIA command line interface to submit job to the BiTIA server.

https://bitia.link

(c) 2022-, Subconscious Compute, https://subcom.tech
"""

import typing as T
import os
import functools

import bitia.helper as bhelper
from bitia.logger import logger

import bitia
import bitia.session as bsession
import bitia.config as bconfig


import typer

app = typer.Typer()


def version_callback(value: bool):
    """callback for version"""
    if value:
        print(bitia.version())


def session(func):
    """Load a session before and save the session after the function call"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        bsession.load_session()
        retval = func(*args, **kwargs)
        bsession.save_session()
        return retval

    return wrapper


@app.command("create-container")
@session
def create_remote_container(user_input: str, recreate: bool = False):
    """Run a pipeline"""
    pipeline_zip, _ = bhelper.user_input_to_pipeline(user_input)
    res = bhelper.post_pipeline_task(
        pipeline_zip,
        endpoint="container/create",
        params=dict(recreate="true" if recreate else "false"),
        server=bconfig.server(),
        stream=True,
    )
    res.raise_for_status()
    for line in res.iter_lines():
        print(line.decode().rstrip())


@app.command("list-container")
@session
def list_remote_container(
    user_input: str, server: str = bconfig.g_server
) -> T.List[str]:
    """List the remote server associated with the pipeline."""
    _, pipeline_hash = bhelper.user_input_to_pipeline(user_input)
    logger.info(f"Finding container for {user_input}, sha256sum={pipeline_hash}")
    res = bhelper.get(
        endpoint="container/list",
        server=bconfig.server(),
        params=dict(pipeline_sha256=pipeline_hash),
    )
    res.raise_for_status()
    _json = res.json()
    print(_json)
    return _json["containers"]


@app.command("logs")
@session
def stream_log(user_input: str, server: str = bconfig.g_server):
    """Stream logs for the most recent run of a given pipeline."""
    _, pipeline_hash = bhelper.user_input_to_pipeline(user_input)
    logger.info(f"Finding container for {user_input}, sha256sum={pipeline_hash}")
    res = bhelper.get(
        endpoint="logs",
        params=dict(pipeline_sha256=pipeline_hash),
        server=bconfig.server(),
        stream=True,
    )
    res.raise_for_status()
    for line in res.iter_lines():
        print(line.decode().rstrip())


@app.command("submit")
@session
def submit_pipeline(user_input: str, rerun: bool = False):
    """Submit your pipelin (url, directory, zip_file).

    Prepare the user directory to send to the server. User can also provide link
    to the pipeline to run.
    """
    create_remote_container(user_input, recreate=rerun)
    containers = list_remote_container(user_input)
    logger.info(f" Got containers {containers}")
    return containers


@app.command("run")
@session
def run_pipeline(user_input: str, *, rerun: bool = False):
    """Run a pipeline"""
    create_remote_container(user_input, recreate=rerun)
    containers = list_remote_container(user_input)
    return [
        bhelper.log_container(container, server=bconfig.server())
        for container in containers
    ]


@app.callback()
def main(verbose: bool = False, server: T.Optional[str] = None):
    """
    Callback
    """
    if verbose:
        logger.info("--verbose mode enabled")

    if server is not None:
        bconfig.set_server(server)
    elif os.environ.get("BITIA_SERVER") is not None:
        bconfig.set_server(os.environ["BITIA_SERVER"])
    else:
        pass
    logger.info(f"Using server {bconfig.server()}")


@app.command()
def version():
    """version information"""
    print(bitia.version())


if __name__ == "__main__":
    app()
